#include "SHERPA/PerturbativePhysics/Perturbative_Interface.H"

#include "SHERPA/PerturbativePhysics/Shower_Handler.H"
#include "SHERPA/PerturbativePhysics/Matrix_Element_Handler.H"
#include "SHERPA/PerturbativePhysics/MI_Handler.H"
#include "SHERPA/SoftPhysics/Hadron_Decay_Handler.H"
#include "PDF/Main/Shower_Base.H"
#include "ATOOLS/Phys/Cluster_Amplitude.H"
#include "PHASIC++/Main/Process_Integrator.H"
#include "PHASIC++/Process/Single_Process.H"
#include "PHASIC++/Process/ME_Generator_Base.H"
#include "PHASIC++/Process/POWHEG_Process.H"
#include "ATOOLS/Org/Exception.H"
#include "ATOOLS/Org/MyStrStream.H"
#include "ATOOLS/Org/Data_Reader.H"
#include "ATOOLS/Math/Random.H"

using namespace SHERPA;
using namespace PHASIC;
using namespace ATOOLS;

Perturbative_Interface::Perturbative_Interface
(Matrix_Element_Handler *const meh,Shower_Handler *const psh):
  p_me(meh), p_mi(NULL), p_hd(NULL), p_shower(psh),
  p_ampl(NULL), m_cmode(0)
{
  Data_Reader read(" ",";","!","=");
  read.AddComment("#");
  read.SetInputPath(p_me->Path());
  read.SetInputFile(p_me->File());
  m_cmode=ToType<int>(rpa.gen.Variable("METS_CLUSTER_MODE"));
  m_bbarmode=read.GetValue<int>("METS_BBAR_MODE",1);
}

Perturbative_Interface::Perturbative_Interface
(MI_Handler *const mi,Shower_Handler *const psh):
  p_me(NULL), p_mi(mi), p_hd(NULL), p_shower(psh),
  p_ampl(NULL), m_cmode(0) {}

Perturbative_Interface::Perturbative_Interface
(Hadron_Decay_Handler *const hdh,Shower_Handler *const psh):
  p_me(NULL), p_mi(NULL), p_hd(hdh), p_shower(psh),
  p_ampl(NULL), m_cmode(0) {}

Perturbative_Interface::~Perturbative_Interface() 
{
  if (p_ampl) p_ampl->Delete();
}

bool Perturbative_Interface::SetColours
(Cluster_Amplitude *ampl,Blob *const bl)
{
  msg_Debugging()<<METHOD<<"(): {\n";
  while (ampl->Prev()) ampl=ampl->Prev();
  for (size_t i(0);i<ampl->Legs().size();++i) {
    Cluster_Leg *cl(ampl->Leg(i));
    if (i<ampl->NIn()) {
      bl->InParticle(i)->SetFlow(1,Max(0,cl->Col().m_j-100));
      bl->InParticle(i)->SetFlow(2,Max(0,cl->Col().m_i-100));
    }
    else {
      bl->OutParticle(i-ampl->NIn())->SetFlow(1,Max(0,cl->Col().m_i-100));
      bl->OutParticle(i-ampl->NIn())->SetFlow(2,Max(0,cl->Col().m_j-100));
    }
  }
  msg_Debugging()<<*bl<<"\n";
  msg_Debugging()<<"}\n";
  return true;
}

Return_Value::code Perturbative_Interface::
DefineInitialConditions(ATOOLS::Blob *blob) 
{
  if (blob==NULL) {
    msg_Error()<<METHOD<<"(): Signal process not found."<<std::endl;
    return Return_Value::Error;
  }
  p_hard=blob;
  if (!p_shower->On()) {
    m_weight=1.0;
    return Return_Value::Success;
  }
  if (p_ampl) {
    p_ampl->Delete();
    p_ampl=NULL;
  }
  p_shower->CleanUp();
  msg_Indent();
  if (p_mi) {
    p_ampl=p_mi->ClusterConfiguration();
    if (!SetColours(p_ampl,blob)) return Return_Value::New_Event;
    if (!p_shower->GetShower()->PrepareShower(p_ampl)) return Return_Value::New_Event;
    return Return_Value::Success;
  }
  if (p_hd) {
    p_ampl=p_hd->ClusterConfiguration(blob);
    if (!SetColours(p_ampl,blob)) return Return_Value::New_Event;
    if (!p_shower->GetShower()->PrepareShower(p_ampl)) return Return_Value::New_Event;
    return Return_Value::Success;
  }
  p_ampl=p_me->Process()->Get<Single_Process>()->Cluster(m_cmode);
  if (p_ampl==NULL) return Return_Value::New_Event;
  size_t cmax(0);
  for (size_t i(0);i<p_ampl->Legs().size();++i)
    cmax=Max(cmax,(size_t)p_ampl->Leg(i)->Col().m_i);
  while (Flow::Counter()<cmax);
  p_me->Process()->Parent()->SetRBMap(p_ampl);
  if (!SetColours(p_ampl,blob)) return Return_Value::New_Event;
  m_weight=1.0;
  if (p_me->Process()->Info().m_ckkw&1) {
    if (m_bbarmode && p_me->HasNLO()==2 &&
        p_me->Process()->Parent()->Info().m_fi.NLOType()==nlo_type::lo) {
      // Bbar reweighting for smooth merging
      DEBUG_FUNC("Bbar reweighting");
      Cluster_Amplitude *ampl=p_ampl->CopyAll();
      if (!LocalKFactor(ampl)) {
        DEBUG_INFO("didn't find PowProc along these cluster amplitudes. "
                   <<"trying with exclusively clustered amplitude:");
        Cluster_Amplitude *excl_ampl=
            p_me->Process()->Get<Single_Process>()->Cluster(m_cmode|256|512);
	if (excl_ampl) {
        if (!LocalKFactor(excl_ampl)) {
          DEBUG_INFO("didn't find PowProc in exclusively clustered amplitude");
	}
        while (excl_ampl->Prev()) excl_ampl=excl_ampl->Prev();
        excl_ampl->Delete();
	}
      }
      while (ampl->Prev()) ampl=ampl->Prev();
      ampl->Delete();
    }
    blob->AddData("Sud_Weight",new Blob_Data<double>(m_weight));
    if (p_me->EventGenerationMode()==1) {
      if (m_weight>=ran.Get()) m_weight=1.0;
      else return Return_Value::New_Event;
    }
    Blob_Data_Base *winfo((*blob)["Weight"]);
    if (!winfo) THROW(fatal_error,"No weight information in signal blob");
    double meweight(winfo->Get<double>());
    blob->AddData("Weight",new Blob_Data<double>(meweight*m_weight));
  }
  if (!p_shower->GetShower()->PrepareShower(p_ampl)) return Return_Value::New_Event;
  return Return_Value::Success;
}

bool Perturbative_Interface::LocalKFactor(ATOOLS::Cluster_Amplitude* ampl)
{
  DEBUG_FUNC(ampl->Legs().size());
  Process_Vector procs(p_me->AllProcesses());
  while (ampl->Next()!=NULL) {
    ampl=ampl->Next();
    Process_Base::SortFlavours(ampl);
    Flavour_Vector amplflavs;
    Vec4D_Vector pb;
    for (size_t i=0; i<ampl->Legs().size(); ++i) {
      amplflavs.push_back(i<ampl->NIn() ?
                          ampl->Leg(i)->Flav().Bar():ampl->Leg(i)->Flav());
      pb.push_back(i<ampl->NIn()?-ampl->Leg(i)->Mom():ampl->Leg(i)->Mom());
    }
    DEBUG_VAR(amplflavs);
    for (size_t i=0; i<procs.size(); ++i) {
      POWHEG_Process* powproc = dynamic_cast<POWHEG_Process*>(procs[i]);
      if (powproc==NULL) continue;
      if (ampl->Legs().size()!=procs[i]->NIn()+procs[i]->NOut()-1) continue;
      Process_Base* bornprocs = (*powproc)[2];
      for (size_t iflav=0; iflav<bornprocs->Size(); ++iflav) {
        DEBUG_VAR((*bornprocs)[iflav]->Flavours());
        if (amplflavs==(*bornprocs)[iflav]->Flavours()) {
          m_weight*=powproc->LocalKFactor(pb);
          DEBUG_VAR(m_weight);
          return true;
        }
      }
    }
  }
  // no process found along ampl
  return false;
}

bool Perturbative_Interface::FillBlobs(ATOOLS::Blob_List *blobs)
{
  if (p_hard==NULL) return false;
  Blob *sblob = new Blob();
  sblob->SetType(btp::Shower);
  sblob->SetStatus(blob_status::needs_showers);
  sblob->SetId();
  if (p_shower->On()) {
    if (!p_hd)
      for (int i(0);i<p_hard->NInP();++i)
	sblob->AddToOutParticles(p_hard->InParticle(i));
    for (int i(0);i<p_hard->NOutP();++i)
      sblob->AddToInParticles(p_hard->OutParticle(i));
  }
  blobs->push_back(sblob);
  p_shower->FillBlobs(blobs); 
  return true;
}

int Perturbative_Interface::PerformShowers()
{
  Blob_Data_Base *winfo((*p_hard)["Weight"]);
  if (!winfo) THROW(fatal_error,"No weight information in signal blob");
  double meweight(winfo->Get<double>());
  if (meweight==0.0) return 0;
  PDF::Shower_Base *csh(p_shower->GetShower());
  int stat=csh->PerformShowers();
  double weight=csh->Weight();
  p_hard->AddData("Shower_Weight",new Blob_Data<double>(weight));
  p_hard->AddData("Weight",new Blob_Data<double>(meweight*weight));
  return stat;
}

int Perturbative_Interface::PerformDecayShowers()
{ 
  return p_shower->GetShower()->PerformDecayShowers(); 
}

void Perturbative_Interface::CleanUp()
{
}

